function log10(x) {
    return Math.log(x) / Math.LN10;
}
export class Sound {
    constructor(buffer) {
        this.isInContext = false;
        this.isPlaying = false;
        this.buffer = buffer;
        this.loop = false;
        this.loopStart = 0;
        this.loopEnd = this.buffer.length;
    }
    setContext(context) {
        this.volume = context.createGain();
        this.volume.gain.setValueAtTime(0.0000001, 0);
        this.panner = context.createPanner();
        this.panner.panningModel = 'HRTF';
        this.panner.distanceModel = 'inverse';
        this.panner.refDistance = 1;
        this.panner.maxDistance = 1000;
        this.panner.rolloffFactor = 1;
        this.panner.coneInnerAngle = 360;
        this.panner.coneOuterAngle = 0;
        this.panner.coneOuterGain = 0;
        this.panner.orientationX.setValueAtTime(0, 0);
        this.panner.orientationY.setValueAtTime(0, 0);
        this.panner.orientationZ.setValueAtTime(-1, 0);
        this.panner.positionX.setValueAtTime(0, 0);
        this.panner.positionY.setValueAtTime(0, 0);
        this.panner.positionZ.setValueAtTime(0, 0);
        this.source = context.createBufferSource();
        this.source.onended = () => {
            this.isPlaying = false;
        };
        this.source.connect(this.volume);
        this.volume.connect(this.panner);
        this.source.buffer = this.buffer;
        this.source.loop = this.loop;
        this.isInContext = true;
    }
    get nodeInput() {
        return this.source;
    }
    get node() {
        return this.panner;
    }
    play(time = 0) {
        if (!this.isInContext) {
            throw new Error('Sound not in context');
        }
        if (this.source && this.loop) {
            this.source.loop = this.loop;
            this.source.loopStart = this.loopStart;
            this.source.loopEnd = this.loopEnd;
            this.source.start(time);
        }
        else if (this.source) {
            this.source.loop = false;
            this.source.start(time);
        }
    }
    stop() {
        if (this.source)
            this.source.stop();
    }
    destroy() {
        var _a, _b, _c;
        (_a = this.source) === null || _a === void 0 ? void 0 : _a.stop();
        if (this.volume)
            (_b = this.source) === null || _b === void 0 ? void 0 : _b.disconnect(this.volume);
        if (this.panner)
            (_c = this.volume) === null || _c === void 0 ? void 0 : _c.disconnect(this.panner);
    }
}
export class SoundSystem {
    constructor() {
        this.reboot();
    }
    static addSoundToContext(context, sound) {
        sound.setContext(context);
    }
    context() {
        return SoundSystem.soundContext;
    }
    masterGainNode() {
        return SoundSystem.masterGainNode;
    }
    get nodeInput() {
        return this.masterGainNode();
    }
    reboot() {
        try {
            const AudioContext = window.webkitAudioContext
                || window.AudioContext
                || false;
            if (AudioContext) {
                if (SoundSystem.soundContext) {
                    SoundSystem.soundContext.suspend();
                }
                SoundSystem.soundContext = new AudioContext();
                SoundSystem.soundContext.onstatechange = this.handleStateChange;
                SoundSystem.soundContext.audioWorklet.addModule('noise.processor.js');
                SoundSystem.analyser = SoundSystem.soundContext.createAnalyser();
                SoundSystem.masterGainNode = SoundSystem.soundContext.createGain();
                SoundSystem.masterGainNode.gain.value = 1;
                SoundSystem.masterGainNode.connect(SoundSystem.soundContext.destination);
                SoundSystem.masterGainNode.connect(SoundSystem.analyser);
                window.addEventListener('focus', (event) => this.resumePlay());
            }
            else {
                console.error('Andy: Web Audio API is not supported in this browser');
            }
        }
        catch (e) {
            console.error('Andy: Error creating Web Audio context', e);
        }
    }
    getAnalyser() {
        return SoundSystem.analyser;
    }
    volume() {
        return SoundSystem.masterGainNode.gain.value;
    }
    volumeInDB() {
        return 20 * log10(this.volume());
    }
    setVolumeInDB(decibel) {
        this.adjustVolume(Math.pow(10, (decibel / 20)));
    }
    adjustVolume(by) {
        SoundSystem.masterGainNode.gain.value = by;
    }
    play(sound, multiple = false, time = 0) {
        if (multiple === true || !sound.isPlaying) {
            this.playSound(sound, time);
        }
    }
    stop(sound) {
        if (sound.isPlaying) {
            sound.stop();
            sound.isPlaying = false;
        }
    }
    currentTime() {
        if (!SoundSystem.soundContext) {
            return 0;
        }
        return SoundSystem.soundContext.currentTime;
    }
    decodeSound(data) {
        return new Promise((res, rej) => {
            let buffer;
            if (SoundSystem.soundContext) {
                SoundSystem.soundContext.decodeAudioData(data, (ab) => {
                    buffer = ab;
                    res(buffer);
                }, (error) => {
                    console.error(error);
                    rej(error);
                });
            }
        });
    }
    resumePlay() {
        return SoundSystem.soundContext.resume();
    }
    playSound(sound, time = 0) {
        var _a;
        sound.setContext(this.context());
        if (sound.node) {
            sound.node.connect(this.masterGainNode());
        }
        (_a = sound.volume) === null || _a === void 0 ? void 0 : _a.gain.setValueAtTime(0.8, time);
        sound.play(time);
        sound.isPlaying = true;
    }
    handleStateChange(e) {
        if (SoundSystem.soundContext && SoundSystem.soundContext.state !== 'running') {
            this.resumePlay();
        }
        return undefined;
    }
}
export const loadSound = (url, soundSystem) => {
    return new Promise((resolve, reject) => {
        const headers = new Headers();
        headers.append('User-Agent', 'Andy/1.0');
        const body = undefined;
        fetch(url, {
            method: 'GET',
            mode: 'cors',
            cache: 'reload',
            headers: headers,
            credentials: 'same-origin',
            redirect: 'follow',
            referrerPolicy: 'no-referrer',
            body: (body) ? JSON.stringify(body) : body,
        })
            .then(response => response.arrayBuffer())
            .then(response => {
            soundSystem.decodeSound(response).then((b) => {
                const sound = new Sound(b);
                resolve(sound);
            }, (err) => {
                reject(err);
            }).catch(e => console.error(e));
        }).catch(e => console.error(e));
    });
};
