import { NoOp } from "../digital/types";
import { MidiToFreq, NoteToMidi } from './types';
export const DURATION_CLICK_PAD = .015;
function createCompressor(system, vpc, _length, time) {
    const compressor = system.context().createDynamicsCompressor();
    compressor.threshold.setValueAtTime(vpc.threshold, time);
    compressor.knee.setValueAtTime(vpc.knee, time);
    compressor.ratio.setValueAtTime(vpc.ratio, time);
    compressor.attack.setValueAtTime(vpc.attack, time);
    compressor.release.setValueAtTime(vpc.release, time);
    return compressor;
}
function createADSR(system, padsr, length, time) {
    const asdr = system.context().createGain();
    const now = time;
    const atkDuration = padsr.attack * length;
    const atkEndTime = now + atkDuration;
    const decayDuration = padsr.decay * length;
    asdr.gain.setValueAtTime(0.00001, now);
    asdr.gain.exponentialRampToValueAtTime(1, atkEndTime);
    asdr.gain.exponentialRampToValueAtTime(padsr.sustain, now + decayDuration);
    const relDuration = padsr.release * length;
    const relEndTime = now + relDuration;
    asdr.gain.exponentialRampToValueAtTime(padsr.sustain, now + relDuration);
    asdr.gain.exponentialRampToValueAtTime(0.00001, now + length + DURATION_CLICK_PAD);
    return asdr;
}
function createFilter(system, pf, _length, time) {
    var _a;
    const filter = system.context().createBiquadFilter();
    filter.type = pf.type;
    filter.frequency.setValueAtTime(pf.frequency, time);
    filter.detune.setValueAtTime((_a = pf === null || pf === void 0 ? void 0 : pf.detune) !== null && _a !== void 0 ? _a : 1, time);
    filter.Q.setValueAtTime(pf.q, time);
    filter.gain.setValueAtTime(pf.gain, time);
    return filter;
}
export function patchPlayer(system, patch) {
    return (note, time, output) => {
        var _a, _b;
        const midi = NoteToMidi.get(note);
        if (!midi) {
            console.warn(`Could not find midi value for note: ${note} Bad note?`);
            return NoOp;
        }
        const fundamental = MidiToFreq[midi];
        if (!fundamental)
            throw Error(`Could not find frequency value for note. Bad midi? ${midi}`);
        let oscillatorOutput;
        const soundLen = patch.len;
        const patchGain = system.context().createGain();
        const sampleGain = (_b = (_a = patch.vca) === null || _a === void 0 ? void 0 : _a.gain) !== null && _b !== void 0 ? _b : 1;
        patchGain.gain.linearRampToValueAtTime(sampleGain, time);
        const processing = new Array();
        if (patch.filter) {
            processing.push(createFilter(system, patch.filter, soundLen, time));
        }
        if (patch.adsr) {
            processing.push(createADSR(system, patch.adsr, soundLen, time));
        }
        if (patch.compressor) {
            processing.push(createCompressor(system, patch.compressor, soundLen, time));
        }
        processing.push(patchGain);
        if (processing.length > 1) {
            for (let z = 1; z < processing.length; z++) {
                const fx = processing[z];
                (processing[z - 1]).connect(fx);
            }
        }
        oscillatorOutput = processing[0];
        const vcoCleanUp = [];
        const voiceLen = Math.min(patch.voices.length, 8);
        for (let v = 0; v < voiceLen; v++) {
            const voice = patch.voices[v];
            let vco;
            if (voice.vco.type === 'custom') {
                vco = new AudioWorkletNode(system.context(), 'white-noise-processor');
            }
            else {
                vco = system.context().createOscillator();
                const lvco = vco;
                const voice = patch.voices[v];
                lvco.type = voice.vco.type;
                if (voice.vco.frequency) {
                    lvco.frequency.value = voice.vco.frequency;
                }
                else if (voice.vco.freqOffset) {
                    lvco.frequency.value = fundamental + voice.vco.freqOffset;
                }
                else if (voice.vco.freqMult) {
                    lvco.frequency.value = fundamental * voice.vco.freqMult;
                }
                else {
                    lvco.frequency.value = fundamental;
                }
                lvco.detune.value = voice.vco.detune;
                lvco.start(time);
                if (soundLen > 0)
                    lvco.stop(time + soundLen + DURATION_CLICK_PAD * 10);
            }
            const vca = system.context().createGain();
            const pvca = patch.voices[v].vca;
            let vcaVolume = 1;
            if (pvca.gainOffset) {
                vcaVolume = 1 + pvca.gainOffset;
            }
            if (pvca.gain) {
                vcaVolume = pvca.gain;
            }
            vca.gain.setValueAtTime(vcaVolume, time);
            vco.connect(vca);
            vca.connect(oscillatorOutput);
            if (soundLen > 0) {
                vca.gain.exponentialRampToValueAtTime(vcaVolume, (time + soundLen - DURATION_CLICK_PAD));
                vca.gain.exponentialRampToValueAtTime(0.00001, (time + soundLen + DURATION_CLICK_PAD));
            }
            vcoCleanUp.push((t) => {
                var _a, _b;
                vca.gain.exponentialRampToValueAtTime(0.00001, t);
                if ((_a = vco) === null || _a === void 0 ? void 0 : _a.stop) {
                    (_b = vco) === null || _b === void 0 ? void 0 : _b.stop();
                }
                vco.disconnect(vca);
                vca.disconnect(oscillatorOutput);
            });
        }
        patchGain.connect(output);
        return (t) => {
            const time = t !== null && t !== void 0 ? t : system.currentTime();
            patchGain.gain.setValueAtTime(patchGain.gain.value, time);
            patchGain.gain.linearRampToValueAtTime(0, time + (DURATION_CLICK_PAD * 22));
            setTimeout(() => {
                vcoCleanUp.forEach((cleanUp) => cleanUp(time + (DURATION_CLICK_PAD * 10)));
                if (processing.length > 1) {
                    for (let z = 1; z < processing.length; z++) {
                        const fx = processing[z];
                        (processing[z - 1]).disconnect(fx);
                    }
                }
                patchGain.disconnect(output);
            }, 333);
        };
    };
}
