import { BinaryFile } from "../io/file";
var ChunkType;
(function (ChunkType) {
    ChunkType["Header"] = "MThd";
    ChunkType["Track"] = "MTrk";
})(ChunkType || (ChunkType = {}));
export var Format;
(function (Format) {
    Format[Format["Single"] = 0] = "Single";
    Format[Format["MultiSync"] = 1] = "MultiSync";
    Format[Format["MultiAsync"] = 2] = "MultiAsync";
})(Format || (Format = {}));
export class MIDI {
    constructor(data) {
        this.markers = new Map([]);
        this.data = new BinaryFile(data);
        this.format = 0;
        this.numberOfTracks = 0;
        this.ticksPerQuarterNote = 0;
        this.tracks = new Array(this.numberOfTracks);
        this.readHeader();
        for (let t = 0; t < this.numberOfTracks; t++) {
            this.tracks[t] = this.readTrack();
        }
    }
    readHeader() {
        const hdr = this.data.readString(4);
        const size = this.data.readUInt32(false);
        if (hdr !== ChunkType.Header || size !== 6) {
            throw new Error(`Incorrect header, not a midi file? ${hdr} ${size}`);
        }
        const format = this.data.readUInt16(false);
        this.format = format;
        const tracks = this.data.readUInt16(false);
        this.numberOfTracks = tracks;
        const ticks = this.data.readUInt16(false);
        if ((ticks & 0x8000) === 0) {
            this.ticksPerQuarterNote = ticks;
        }
        else {
            throw new Error("Divisions set as units per SMTPE frame: Unsupported");
        }
    }
    readTrack() {
        const events = [];
        const track = {
            name: "Unknown",
            events,
        };
        const hdr = this.data.readString(4);
        const length = this.data.readUInt32(false);
        const start = this.data.tell();
        let lastEvent = 0;
        let trackTime = 0;
        while (true) {
            let dtime = this.readVariableLen();
            let eventType = this.data.readUInt8();
            trackTime += dtime;
            if (this.data.tell() >= (start + length)) {
                break;
            }
            if ((eventType & 0xF0) === 0xF0) {
                if (eventType === 0xFF) {
                    const event = this.metaEvents(dtime, trackTime, eventType, track);
                    events.push(event);
                    if (event.raw === 0x2F) {
                        break;
                    }
                }
                else if (eventType === 0xF0 || eventType === 0xF7) {
                    const event = this.sysexEvents(dtime, trackTime, eventType);
                    events.push(event);
                }
                else {
                    console.warn(`Unknown event ${eventType.toString(16)}`);
                }
            }
            else {
                const event = this.midiEvents(dtime, trackTime, eventType, lastEvent);
                lastEvent = event.raw;
                events.push(event);
            }
        }
        return track;
    }
    metaEvents(dtime, ttime, eventType, track) {
        var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k, _l, _m, _o, _p, _q, _r, _s, _t, _u;
        const event = {
            raw: eventType,
            time: dtime,
            abstime: ttime,
            type: 'Meta',
            data: [],
        };
        const command = this.data.readUInt8();
        const length = this.readVariableLen();
        event.raw = command;
        switch (command) {
            case 0x00:
                {
                    event.subtype = 'SequenceNumber';
                    (_a = event.data) === null || _a === void 0 ? void 0 : _a.push(this.data.readUInt16(false));
                }
                break;
            case 0x01:
                {
                    event.subtype = 'Text';
                    const txt = this.data.readString(length);
                    (_b = event.data) === null || _b === void 0 ? void 0 : _b.push(txt);
                }
                break;
            case 0x02:
                {
                    event.subtype = 'CopyrightNotice';
                    const copy = this.data.readString(length);
                    (_c = event.data) === null || _c === void 0 ? void 0 : _c.push(copy);
                }
                break;
            case 0x03:
                {
                    event.subtype = 'TrackName';
                    const name = this.data.readString(length);
                    (_d = event.data) === null || _d === void 0 ? void 0 : _d.push(name);
                    track.name = name;
                }
                break;
            case 0x04:
                {
                    event.subtype = 'InstrumentName';
                    const name = this.data.readString(length);
                    (_e = event.data) === null || _e === void 0 ? void 0 : _e.push(name);
                }
                break;
            case 0x05:
                {
                    event.subtype = 'Lyrics';
                    const lyrics = this.data.readString(length);
                    (_f = event.data) === null || _f === void 0 ? void 0 : _f.push(lyrics);
                }
                break;
            case 0x06:
                {
                    event.subtype = 'Marker';
                    const markerStr = this.data.readString(length);
                    (_g = event.data) === null || _g === void 0 ? void 0 : _g.push(markerStr);
                    this.markers.set(markerStr, { time: dtime, abstime: ttime });
                }
                break;
            case 0x07:
                {
                    event.subtype = 'CuePoint';
                    const cueStr = this.data.readString(length);
                    (_h = event.data) === null || _h === void 0 ? void 0 : _h.push(cueStr);
                }
                break;
            case 0x0f:
                {
                    event.subtype = 'Unknown';
                    if (length > 0) {
                        const str = this.data.readBytes(length);
                        (_j = event.data) === null || _j === void 0 ? void 0 : _j.push(str);
                    }
                }
                break;
            case 0x20:
                {
                    event.subtype = 'MidiChannelPrefix';
                    (_k = event.data) === null || _k === void 0 ? void 0 : _k.push(this.data.readUInt8());
                }
                break;
            case 0x21:
                {
                    event.subtype = 'Unknown';
                    (_l = event.data) === null || _l === void 0 ? void 0 : _l.push(this.data.readUInt8());
                }
                break;
            case 0x2f:
                {
                    event.subtype = 'EndOfTrack';
                }
                break;
            case 0x51:
                {
                    event.subtype = 'SetTempo';
                    const microseconds = (this.data.readUInt8() << 16) +
                        (this.data.readUInt8() << 8) +
                        this.data.readUInt8();
                    (_m = event.data) === null || _m === void 0 ? void 0 : _m.push(microseconds);
                }
                break;
            case 0x54:
                {
                    event.subtype = 'SMPTEOffset';
                    if (length > 0)
                        this.data.readBytes(length);
                }
                break;
            case 0x58:
                {
                    event.subtype = 'TimeSignature';
                    (_o = event.data) === null || _o === void 0 ? void 0 : _o.push(this.data.readUInt8());
                    (_p = event.data) === null || _p === void 0 ? void 0 : _p.push(Math.pow(2, this.data.readUInt8()));
                    (_q = event.data) === null || _q === void 0 ? void 0 : _q.push(this.data.readUInt8());
                    (_r = event.data) === null || _r === void 0 ? void 0 : _r.push(this.data.readUInt8());
                }
                break;
            case 0x59:
                {
                    event.subtype = 'KeySignature';
                    (_s = event.data) === null || _s === void 0 ? void 0 : _s.push(this.data.readInt8());
                    (_t = event.data) === null || _t === void 0 ? void 0 : _t.push(this.data.readUInt8());
                }
                break;
            case 0x7F:
                {
                    event.subtype = "SequencerSpecific";
                    const _data = this.data.readBytes(length);
                }
                break;
            default: {
                console.warn(`Unknown meta event ${eventType.toString(16)} ${command.toString(16)} ${length}`);
                event.subtype = 'Unknown';
                if (length > 0)
                    (_u = event.data) === null || _u === void 0 ? void 0 : _u.push(this.data.readBytes(length));
            }
        }
        return event;
    }
    sysexEvents(dtime, ttime, eventType) {
        const event = {
            raw: eventType,
            abstime: ttime,
            time: dtime,
            type: 'SysEx',
            data: [],
        };
        const length = this.readVariableLen();
        const _data = this.data.readBytes(length);
        return event;
    }
    midiEvents(dtime, ttime, command, lastEvent) {
        var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k, _l;
        let useEvent = 0;
        let param = 0;
        if ((command & 0x80) === 0) {
            param = command;
            useEvent = lastEvent;
        }
        else {
            param = this.data.readUInt8();
            useEvent = command;
        }
        const event = {
            raw: useEvent,
            time: dtime,
            abstime: ttime,
            type: 'Channel',
            data: [],
        };
        const cmd = ((useEvent & 0xF0) >> 4);
        const channel = ((useEvent & 0x0F));
        event.channel = channel;
        switch (cmd) {
            case 0x8:
                {
                    event.subtype = 'NoteOff';
                    const note = param;
                    (_a = event.data) === null || _a === void 0 ? void 0 : _a.push(note);
                    const velocity = this.data.readUInt8();
                    (_b = event.data) === null || _b === void 0 ? void 0 : _b.push(velocity);
                }
                break;
            case 0x9:
                {
                    const note = param;
                    (_c = event.data) === null || _c === void 0 ? void 0 : _c.push(note);
                    const velocity = this.data.readUInt8();
                    (_d = event.data) === null || _d === void 0 ? void 0 : _d.push(velocity);
                    if (velocity == 0)
                        event.subtype = 'NoteOff';
                    else
                        event.subtype = 'NoteOn';
                }
                break;
            case 0xa:
                {
                    event.subtype = 'NoteAfterTouch';
                    (_e = event.data) === null || _e === void 0 ? void 0 : _e.push(param);
                    const amount = this.data.readUInt8();
                    (_f = event.data) === null || _f === void 0 ? void 0 : _f.push(amount);
                }
                break;
            case 0xb:
                {
                    event.subtype = 'Controller';
                    (_g = event.data) === null || _g === void 0 ? void 0 : _g.push(param);
                    const value = this.data.readUInt8();
                    (_h = event.data) === null || _h === void 0 ? void 0 : _h.push(value);
                }
                break;
            case 0xc:
                {
                    event.subtype = 'ProgramChange';
                    const type = param;
                    (_j = event.data) === null || _j === void 0 ? void 0 : _j.push(type);
                }
                break;
            case 0xd:
                {
                    event.subtype = 'ChannelAfterTouch';
                    (_k = event.data) === null || _k === void 0 ? void 0 : _k.push(param);
                }
                break;
            case 0xe:
                {
                    event.subtype = 'PitchBend';
                    const amount = param + (this.data.readUInt8() << 7);
                    (_l = event.data) === null || _l === void 0 ? void 0 : _l.push(amount);
                }
                break;
            default:
                console.warn(`[[UNKNOWN ${command.toString(16)}]]`);
        }
        return event;
    }
    readVariableLen() {
        const start = this.data.readUInt8();
        if ((start & 0x80) === 0) {
            return start;
        }
        let c = 0;
        let value = start & 0x7F;
        do {
            value = (value << 7) + ((c = this.data.readUInt8() & 0x7F));
        } while (c & 0);
        return value;
    }
}
export const loadMidi = (url) => {
    return new Promise((resolve, reject) => {
        const headers = new Headers();
        headers.append('User-Agent', 'Andy/1.0');
        headers.append('Accept', 'audio/midi');
        const body = undefined;
        fetch(url, {
            method: 'GET',
            mode: 'cors',
            cache: 'reload',
            headers: headers,
            credentials: 'same-origin',
            redirect: 'follow',
            referrerPolicy: 'no-referrer',
            body: (body) ? JSON.stringify(body) : body,
        })
            .then(response => response.arrayBuffer())
            .then(response => {
            const mid = new MIDI(response);
            resolve(mid);
        }).catch(e => console.error(e));
    });
};
